using Printf

abstract type AbstractMesh end
mutable struct Mesh2dTriangle <: AbstractMesh
	numberofvertices::Int64
	numberofelements::Int64
	x::Vector{Float64}
	y::Vector{Float64}
	elements::Matrix{Int64}
	segments::Matrix{Int64}
	vertexonboundary::Vector{Bool}
end
function Mesh2dTriangle() #{{{
	return Mesh2dTriangle( 0, 0, Vector{Float64}(undef,0), Vector{Float64}(undef, 0), Matrix{Int64}(undef, 0, 0), Matrix{Int64}(undef, 0, 0), Vector{Bool}(undef,0))
end# }}}
function Base.show(io::IO, this::Mesh2dTriangle)# {{{
	IssmStructDisp(io, this)
end# }}}
mutable struct Mesh3dPrism{T} <: AbstractMesh
	numberofvertices::Int64
	numberofelements::Int64
	numberoflayers::Int64
	x::Vector{Float64}
	y::Vector{Float64}
	z::Vector{Float64}
	elements::Matrix{Int64}
	segments::Matrix{Int64}
	vertexonboundary::Vector{Bool}
end
function Mesh3dPrism() #{{{
	return Mesh3dPrism( 0, 0, 0, Vector{Float64}(undef,0), Vector{Float64}(undef,0), Vector{Float64}(undef,0), Matrix{Int64}(undef, 0, 0), Matrix{Int64}(undef, 0, 0), Vector{Bool}(undef,0))
end# }}}

mutable struct Geometry
	surface::Vector{Float64}
	base::Vector{Float64}
	thickness::Vector{Float64}
	bed::Vector{Float64}
end
function Geometry() #{{{
	return Geometry( Vector{Float64}(undef,0), Vector{Float64}(undef,0), Vector{Float64}(undef,0), Vector{Float64}(undef,0))
end# }}}

mutable struct Mask
	ocean_levelset::Vector{Float64}
	ice_levelset::Vector{Float64}
end
function Mask() #{{{
	return Mask( Vector{Float64}(undef,0), Vector{Float64}(undef,0))
end# }}}
function Base.show(io::IO, this::Mask)# {{{
	IssmStructDisp(io, this)
end# }}}

mutable struct Initialization
	vx::Vector{Float64}
	vy::Vector{Float64}
end
function Initialization() #{{{
	return Initialization( Vector{Float64}(undef,0), Vector{Float64}(undef,0))
end# }}}

mutable struct Stressbalance
	spcvx::Vector{Float64}
	spcvy::Vector{Float64}
	restol::Float64
	reltol::Float64
	abstol::Float64
	maxiter::Int64
end
function Stressbalance() #{{{
	return Stressbalance( Vector{Float64}(undef,0), Vector{Float64}(undef,0), 1.e-4, 0.01, 10., 100)
end# }}}
function Base.show(io::IO, this::Stressbalance)# {{{
	IssmStructDisp(io, this)
end# }}}

mutable struct Constants
	g::Float64
	yts::Float64
end
function Constants() #{{{
	return Constants( 9.81,  365*24*3600.)
end# }}}

mutable struct Materials
	rho_ice::Float64
	rho_water::Float64
	rho_freshwater::Float64
	mu_water::Float64
	heatcapacity::Float64
	latentheat::Float64
	thermalconductivity::Float64
	temperateiceconductivity::Float64
	effectiveconductivity_averaging::Int64
	meltingpoint::Float64
	beta::Float64
	mixed_layer_capacity::Float64
	thermal_exchange_velocity::Float64
	rheology_B::Vector{Float64}
	rheology_n::Vector{Float64}
	rheology_law::String
end
function Materials() #{{{
	return Materials(917., 1023., 1000., 0.001787, 2093., 3.34*10^5, 2.4, .24, 1, 273.15, 9.8*10^-8, 3974., 1.00*10^-4, Vector{Float64}(undef,0), Vector{Float64}(undef,0), "Cuffey")
end# }}}

abstract type AbstractFriction end
mutable struct BuddFriction <: AbstractFriction
	coefficient::Vector{Float64}
end
function BuddFriction() #{{{
	return BuddFriction(Vector{Float64}(undef,0))
end# }}}
function Base.show(io::IO, this::BuddFriction)# {{{
	IssmStructDisp(io, this)
end# }}}
mutable struct WeertmanFriction <: AbstractFriction
	C::Vector{Float64}
	m::Vector{Float64}
end

mutable struct model
	mesh::AbstractMesh
	geometry::Geometry
	mask::Mask
	materials::Materials
	initialization::Initialization
	stressbalance::Stressbalance
	constants::Constants
	results
	friction::AbstractFriction
end
function model() #{{{
	return model( Mesh2dTriangle(), Geometry(), Mask(), Materials(), Initialization(),Stressbalance(), Constants(), [], BuddFriction())
end#}}}
function model(matmd::Dict) #{{{

	#initialize output
	md = model()

	#Loop over all possible fields
	for name1 in fieldnames(model)
		if haskey(matmd, String(name1))
			mdfield  = getfield(md,name1)
			matfield = matmd[String(name1)]
			for name2 in fieldnames(typeof(mdfield))
				if haskey(matfield, String(name2))
					println("recovering md.",name1,".",name2)
					value_matlab = matfield[String(name2)]
					value_julia  = getfield(mdfield, name2)

					if typeof(value_matlab)==typeof(value_julia)
						#setfield!(mdfield, name2, convert(typeof(getfield(mdfield, name2)), matfield[String(name2)]))
						setfield!(mdfield, name2, value_matlab)

					elseif typeof(value_matlab)==Float64 && typeof(value_julia)==Int64
						setfield!(mdfield, name2, Int64(value_matlab))

					elseif typeof(value_matlab)==Matrix{Float64} && typeof(value_julia)==Vector{Float64}
						if(size(value_matlab,2)!=1) error("only one column expected") end
						setfield!(mdfield, name2, value_matlab[:,1])

					elseif typeof(value_matlab)==Matrix{Float64} && typeof(value_julia)==Matrix{Int64}
						matrix = Matrix{Int64}(undef,size(value_matlab))
						for i in 1:length(value_matlab) matrix[i] = Int64(value_matlab[i]) end
						setfield!(mdfield, name2, matrix)

					elseif typeof(value_matlab)==Matrix{Float64} && typeof(value_julia)==Vector{Bool}
						if(size(value_matlab,2)!=1) error("only one column expected") end
						vector = Vector{Bool}(undef,size(value_matlab,1))
						for i in 1:length(vector) vector[i] = Bool(value_matlab[i]) end
						setfield!(mdfield, name2, vector)

					else
						error("Don't know how to convert ",typeof(value_matlab)," to ",typeof(value_julia))
					end
				end
			end
			#setfield!(md, name1, mdfield)
		end
	end

	return md
end#}}}
function Base.show(io::IO, md::model)# {{{

	compact = get(io, :compact, false)

	println(io,"Model:")
	@printf "%19s: %-26s -- %s\n" "mesh" typeof(md.mesh) "mesh properties"
	@printf "%19s: %-26s -- %s\n" "geometry" typeof(md.geometry) "surface elevation, bedrock topography, ice thickness,..."
	@printf "%19s: %-26s -- %s\n" "mask" typeof(md.mask) "defines grounded and floating regions"
	@printf "%19s: %-26s -- %s\n" "materials" typeof(md.materials) "material properties"
	@printf "%19s: %-26s -- %s\n" "initialization" typeof(md.initialization) "initial state"
	@printf "%19s: %-26s -- %s\n" "stressbalance" typeof(md.stressbalance) "stress balance parameters"
	@printf "%19s: %-26s -- %s\n" "constants" typeof(md.constants) "physical constants"
	@printf "%19s: %-26s -- %s\n" "friction" typeof(md.friction) "basal friction"

end# }}}
